from tkinter import ttk
from thonny.config_ui import ConfigurationPage
from ..properties import PLUGIN_NAME
from thonny import get_workbench
from ..i18n.languages import tr
from thonny import ui_utils
import tkinter as tk

# Default config
DEFAULT_DOC_GENERATION_AFTER_RETURN = True
DEFAULT_IMPORT_MODULE_IN_SHELL = True
DEFAULT_CLOSE_FUNCTION_ROWS = False
DEFAULT_OPEN_ONLY_RED_FUNCTIONS = True
DEFAULT_HIGHLIGHT_EXCEPTIONS = False
DEFAULT_REPORT_EXCEPTION_DETAIL = True 
DEFAULT_MOVE_ERRORVIEW_TO_BOTTOM = False

# Option names
AUTO_GENERATON_DOC = "auto_generaton_doc"
IMPORT_MODULE = "import_module"
FOLD_ALL_FUNCTIONS = "close_function_rows"
EXPAND_ONLY_RED_FUNCTIONS = "open_only_red_functions"
HIGHLIGHT_EXCEPTIONS = "highlight_exceptions"
REPORT_EXCEPTION_DETAIL = "exception_detail"
MOVE_ERRORVIEW_TO_BOTTOM = "move_errorview_to_bottom"

# Dict of options name and default value
OPTIONS = {
    AUTO_GENERATON_DOC : DEFAULT_DOC_GENERATION_AFTER_RETURN,
    IMPORT_MODULE : DEFAULT_IMPORT_MODULE_IN_SHELL,
    FOLD_ALL_FUNCTIONS: DEFAULT_CLOSE_FUNCTION_ROWS,
    EXPAND_ONLY_RED_FUNCTIONS : not DEFAULT_CLOSE_FUNCTION_ROWS,
    HIGHLIGHT_EXCEPTIONS: DEFAULT_HIGHLIGHT_EXCEPTIONS,
    REPORT_EXCEPTION_DETAIL: DEFAULT_REPORT_EXCEPTION_DETAIL,
    MOVE_ERRORVIEW_TO_BOTTOM: DEFAULT_MOVE_ERRORVIEW_TO_BOTTOM
}

def init_options():
    """
    Initialise dans le workbench les options du plugin.
    """
    for opt in OPTIONS :
        if not get_workbench().get_option(opt) :
            get_workbench().set_default("%s." % PLUGIN_NAME + opt, OPTIONS[opt])

def get_option(name: str): 
    """
    Renvoie la valeur dans le workbench de l'option passée en paramètre.

    Paramètres:
    - name : le nom de l'option, tel que définie ds globals.py 
    """
    return get_workbench().get_option("%s." % PLUGIN_NAME + name)

def set_option(name, value):
    get_workbench().set_option("%s." % PLUGIN_NAME + name, value)

class L1TestConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, AUTO_GENERATON_DOC), 
                          tr("Generate the docstring automatically after a line break at a function name."))

        self.add_checkbox("%s.%s" % (PLUGIN_NAME, IMPORT_MODULE), 
                          tr("Import the module executed in the shell."))
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, FOLD_ALL_FUNCTIONS),
                          tr("Fold all function in %s view.") % PLUGIN_NAME,
                          callback=lambda: set_option(EXPAND_ONLY_RED_FUNCTIONS, not get_option(FOLD_ALL_FUNCTIONS)))
         
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, EXPAND_ONLY_RED_FUNCTIONS),
                          tr("Expand only red functions in %s view.") % PLUGIN_NAME,
                          callback=lambda: set_option(FOLD_ALL_FUNCTIONS, not get_option(EXPAND_ONLY_RED_FUNCTIONS)))
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, HIGHLIGHT_EXCEPTIONS), 
                          tr("Highlight failed tests (only those that throw an exception)."))
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, MOVE_ERRORVIEW_TO_BOTTOM), 
                          tr("Place the error view at the bottom of the Thonny IDE (next to the `Shell` view). " + 
                             "By default, the error view is placed below the `L1Test` view. Thonny must be restarted after having modified this option."))

    def add_checkbox(
            self, flag_name, description, callback=None, row=None, column=0, padx=0, pady=0, columnspan=1, tooltip=None
    ):
        variable = get_workbench().get_variable(flag_name)
        checkbox = ttk.Checkbutton(self, text=description, variable=variable, command=callback)
        checkbox.grid(
            row=row, column=column, sticky=tk.W, padx=padx, pady=pady, columnspan=columnspan
        )

        if tooltip is not None:
            ui_utils.create_tooltip(checkbox, tooltip)



import gettext, os

BASE_LANGUAGE_CODE = "en_US"
BASE_LANGUAGE_NAME = "English"

# https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes

LANGUAGES_DICT = {
    BASE_LANGUAGE_CODE: BASE_LANGUAGE_NAME,
    "fr_FR": "Français",
}


_translation = gettext.NullTranslations()


def tr(message: str) -> str:
    """ Translates the message to the current language."""
    return _translation.gettext(message)


def set_language(language_code: str) -> None:
    global _translation
    try:
        path = os.path.join(os.path.dirname(__file__), "locale")
        _translation = gettext.translation("l1test", path, [language_code])
    except Exception as e:
        _translation = gettext.NullTranslations()


from thonny.editors import EditorCodeViewText
import re
from thonny import get_workbench

# The regex to check if the __main__ already exists
# The regex checks if a string already contains the __main__ and not commented.
_REGEX = re.compile(r"(?m)^\s*if\s+__name__\s*==\s*['\"]__main__['\"]\s*:\s*")

# The generated `__main__` function. 
# Leave a blank line before the `if __name__ == '__main__':` line, 
# so that the generated main function is separated from the rest of the code.
MAIN_TO_GENERATE:str = (
"""

if __name__ == '__main__':
    # éxécuté qd ce module n'est pas initialisé par un import.
    pass\
"""
)

class MainGenerator(): 
    """
    This class is responsible for generating the __main__ function.
    It takes a source code as input and generates the __main__ function 
    at the bottom of the source code.
    
    Args:
        source (str, optional): The source code in which 
        the __main__ will be generated. Defaults to "".
    """         
    def __init__(self, source=""):
        self._source = source
    
    def generate(self, text_widget:EditorCodeViewText=None) -> bool: 
        """
        Generate the __main__ and insert it at the bottom of the given `text_widget`.

        Args:
            text_widget (EditorCodeViewText, optional): The text widget to insert 
            the generated __main__ into. Set to `None` to not insert the generated 
            main function, this is useful for testing.
            
        Returns:
            bool: Returns True if the __main__ is generated, False otherwise.
        """      
        is_generated = False  

        main_line = self.__find_main_lineno(self._source) # Check if a __main__ already exists
        main_content = self.__get_main_content()
        
        if not main_line: # if not found
            is_generated = True
            if text_widget :  
                text_widget.insert("end", main_content)
                main_line = int(text_widget.index('end').split('.')[0]) - (main_content.count("\n") -1)
        if text_widget:
            text_widget.see(f"{main_line}.0")
            end_line = main_line + main_content.count("\n") - 1
            text_widget.select_lines(main_line, end_line if is_generated else main_line) # highlight the generated lines
            get_workbench().after(800, lambda: text_widget.select_lines(0,0)) # remove highlighting the lines after 800ms
        return is_generated
    
    def __find_main_lineno(self, text: str) -> int:
        """
        Returns the line numbers where the __main__ (and not commented) exists. Returns None if not found.
        """
        lines = text.splitlines()
        # use next() to get the first line number where the regex is found. Gets None if not found
        return next((i for i, line in enumerate(lines, 1) if re.search(_REGEX, line)), None)
    
    def __get_main_content(self) -> str:
        """ Returns the content of the __main__ to generate """
        return MAIN_TO_GENERATE

    def get_source(self) -> str:
        """ Returns the source code """
        return self._source
    
    def set_source(self, source: str):
        """ Sets the source code """
        self._source = source


from thonnycontrib.i18n.languages import tr
from thonny import get_workbench
import thonnycontrib.i18n.languages as languages

if get_workbench(): # si l'éditeur est prêt alors on change la langue du translateur.
    languages.set_language(get_workbench().get_option("general.language"))

PLUGIN_NAME = "L1Test"
ERROR_VIEW_LABEL = '%s errors' % PLUGIN_NAME

# ici vous pouvez changer la syntaxe du doctest. 
# version 2022 PJI : Actuellement on garde la syntaxe `$py`.
L1TEST_SYMBOL = "[$]py"
# version 2022 avant la rentrée
L1TEST_SYMBOL1 = "[$][$][$]"
L1TEST_SYMBOL2 = "[$]PY"
L1TEST_SYMBOL3 = "[$]py"
# L'invite des tests qui vérifient la levée d'exception
L1TEST_EXCEPTION_SYMBOL = "[$][$]e"


# ############################################################################################### #
#                       LES NOUVELLES VARIABLES VERSION 2023 PFE                                  #
# ############################################################################################### #

# Le nom de la commande magique pour l1test(doit toujours commencer par une majuscule)
BACKEND_COMMAND = "L1test"


# ############ Les noms des clés du dictionnaire renvoyé par le l1test_backend ############
# Le nom de l'attribut contenant les résulats des tests renvoyés par l1test_backend
VERDICTS = "verdicts"
# Le nom de l'attribut contenant une exception levée et renvoyée par l1test_backend
L1TEST_EXCEPTION = "l1test_exception"

# ############ Les labels des buttons du menu l1test treeview ############
PLACE_RED_TEST_ON_TOP_LABEL = tr("Place the red tests on the top")
RESUME_ORIGINAL_ORDER = tr("Resume original order")
SHOW_ONLY_RED_TESTS = tr("Show only red tests")
SHOW_ALL_TESTS = tr("Show all the tests")
EXPAND_ALL = tr("Expand all functions")
FOLD_ALL = tr("Fold all functions")
UPDATE_FONT_LABEL = tr("Update the font")
INCREASE_SPACE_BETWEEN_ROWS = tr("Inrease row height")
DECREASE_SPACE_BETWEEN_ROWS = tr("Decrease row height")
CLEAR = tr("Clear")

# Le message affiché sur la treeview quand `l1test` est en cours d'execution
L1TEST_IN_PROGRESS = tr("Executing tests in progress")

# Le message affiché sur la treeview quand il n'existe aucun test
NO_TEST_FOUND_MSG = tr("No test found !")

# The title of the error view when the docstring genertor shows the raised error
CANNOT_GENERATE_THE_DOCSTRING = tr("Cannot generate the docstring :")
# The title of the error view when the l1test shows the raised error
CANNOT_RUN_TESTS_MSG = tr("Cannot run %s :")%(PLUGIN_NAME)

# These are the states sent by the evaluator to the L1TestRunner to indicate whether or not test execution has been completed.
PENDING_STATE = "Pending"  # this state indicates that an Example (a test) is still evaluating. This state is sent for each Example (as test).
FINISHED_STATE = "Finished" # this state indicates that the Evaluator has finished it's job and all the evalutions are done. This state is sent after all the evaluation.

# A special event that `L1TestTreeview` sends to `L1TestRunner` when clicking on an exception test
# The event transfers the details of the clicked exception to the `L1TestRunner` 
# which will show it in the error view.
L1TREE_VIEW_EVENT = "L1TreeviewEvent"

# Les images utilisées par la treeview
PENDING_ICON = "pending_icon.png"
ERROR_ICON = "error_icon.png"
RESTART_ICON = "restart_icon.png"
FAILED_RED_CHIP = "failed_red_chip.png" # le petit cercle rouge qui précède un test qui a échoué
EXCEPTION_RED_CHIP = "exception_red_chip.png" # le petit cercle rouge (avec un poitn d'exclamation) qui précède un test qui a échoué
SUCCESS_GREEN_CHIP = "success_green_chip.png" # le petit cercle vert qui précède un test qui a réussi



from abc import *
import ast
from typing import List

class DocTemplate(ABC):
    # Ces constantes peuvent être utilisées dans les classes d'implémention
    NEW_LINE = "\n"
    DOCSTRING_SYMBOL = '"""'
    
    # Les mots utilisés dans le template par défault
    TODO_LABEL = "" #"__à compléter__"
    
    SUMMARY_LABEL = "à_remplacer_par_ce_que_fait_la_fonction"
    PARAM_LABEL = "Paramètres :"
    # nouveauté portail MI : on parle vraiment de précondition !
    CU_LABEL = "Précondition : "
    DOCTEST_LABEL = "Exemple(s) :\n$$$ "
    
    def _format_params(self, params) -> str:
        """
        Args:
            params (List): It's a list of the arguments.

        Returns:
            str: Returns the parameter representation section of a node in a docstring. 
        """
        if params is None:
            return ""
        args_to_exclude = ["self", "cls"]
        label = self.PARAM_LABEL + self.NEW_LINE
        format_params = ""
        for p in params:
            arg_type = ast.unparse(p.annotation) if p.annotation else ""     
            arg_name = p.arg 
            if arg_name not in args_to_exclude: 
                format_params += "- %s (%s) : %s\n" %(arg_name, arg_type, self.TODO_LABEL)
        return label + format_params
    
    @abstractmethod
    def get_parameters(self, node:ast.AST) -> List:
        """
        Get the paramters of a given node.
        
        Args:
            node (ast.AST): An AST node. 

        Returns:
            List: Returns a List of arguments of the given node.
        """
        for node in node.body:
            if isinstance(node, ast.FunctionDef) and node.name == "__init__":
                # Trouver les paramètres de la méthode __init__
                return node.args.args
        return []
    
    @abstractmethod
    def _format_general_summary(self) -> str:
        """
        Returns:
            str: Returns a label which will indicate to write a summary of the function.
        """
        pass
    
    @abstractmethod
    def _format_usage_constraints(self) -> str:
        """
        Returns:
            str: Returns the usage constraints representation section in a docstring.
        """
        pass
    
    @abstractmethod
    def _format_return_value(self) -> str:
        """
        Returns:
            str: Returns the return value representation section in a docstring.
        """
        pass
    
    @abstractmethod
    def _format_test_examples(self) -> str:
        """
        Returns:
            str: Returns the test examples representation section in a docstring.
        """
        pass
    
    @abstractmethod
    def get_template(self, node:ast.AST=None) -> str:
        """Build the complete docstring template. 
        This method must invoke the above abstract methods.
        
        Args:
            node (ast.AST): The AST node in which the dosctring will be generated.

        Returns:
            str: Returns the template representation. 
        """
        pass  
    
    @abstractmethod
    def get_id_signature(self) -> str: 
        pass      
    
class DocFunctionTemplate(DocTemplate):
    '''Modifié pour coller au cours de PROG, portail MI, avec volonté
    d'alléger les docstring au max : uniquement la première phrase, la
    precond et les tests (les étudiant·es étant obligés d'indiquer des
    annotations de type).

    '''
    RETURN_LABEL = "Valeur de retour " 
    RETURN_TYPE_LABEL = "(%s) :" #"__type de retour ?__ (%s)%s"
    
    def get_parameters(self, node:ast.AST):
        """
        Get the paramters of a given node.
        
        Args:
            node (ast.AST): An AST node. Must be an ast.FunctionDef or ast.AsyncFunctionDef

        Returns:
            List: Returns a List of arguments of the given node.
        """
        if isinstance(node, ast.FunctionDef):
            return node.args.args
        return []

    def _format_general_summary(self):
        return self.SUMMARY_LABEL + self.NEW_LINE
    
    def _format_usage_constraints(self):
        return self.CU_LABEL + self.TODO_LABEL + self.NEW_LINE   

    def _format_return_value(self, node: ast):
        return_type = ast.unparse(node.returns) if node.returns else ""
        return_descr = self.RETURN_TYPE_LABEL % return_type + self.NEW_LINE
        return self.RETURN_LABEL + return_descr
    
    def _format_test_examples(self):
        label = self.DOCTEST_LABEL + self.NEW_LINE
        todo = self.TODO_LABEL + self.NEW_LINE
        return label + todo
        
    def get_template(self, node: ast.AST):
        '''Les commentaires indiquent les allègements pour le passage SESI ->
        MI.
        '''
        return (
            self.DOCSTRING_SYMBOL + 
            self._format_general_summary() + self.NEW_LINE + 
            # self._format_params(self.get_parameters(node))  + 
            # self._format_return_value(node) + 
            self._format_usage_constraints() +
            self._format_test_examples() + 
            self.DOCSTRING_SYMBOL + self.NEW_LINE
        )
    
    def get_id_signature(self): 
        return "def" 

class DocClassTemplate(DocTemplate): 
    def _format_general_summary(self):
        return self.SUMMARY_LABEL + self.NEW_LINE
    
    def get_parameters(self, node):
        # Parcourir les définitions de méthodes dans la classe
        for sub_node in node.body:
            if isinstance(sub_node, ast.FunctionDef) and sub_node.name == "__init__":
                # Trouver les paramètres de la méthode __init__
                return sub_node.args.args
        return []
     
    def _format_usage_constraints(self):
        return (self.CU_LABEL +
                self.TODO_LABEL +
                self.NEW_LINE 
            )  

    def _format_return_value(self):
        return ""
    
    def _format_test_examples(self):
        label = self.DOCTEST_LABEL + self.NEW_LINE
        todo = self.TODO_LABEL + self.NEW_LINE
        return label + todo
            
    def get_template(self, node):
        return self.DOCSTRING_SYMBOL + \
               self._format_params(self.get_parameters(node)) + \
               self._format_general_summary() + self.NEW_LINE + \
               self._format_usage_constraints() + \
               self._format_test_examples() + \
               self.DOCSTRING_SYMBOL + self.NEW_LINE

    def get_id_signature(self): 
        return "class" 

class DocTemplateFactory:            
    @staticmethod
    def create_template(type:str):
        return DocTemplateFactory.__search_type(type)
    
    @staticmethod
    def __docTemplate_subclasses(cls=DocTemplate):
        return set(cls.__subclasses__()) \
               .union([s for c in cls.__subclasses__() \
                            for s in DocTemplateFactory.__docTemplate_subclasses(c)])
    
    @staticmethod
    def __search_type(type:str) -> DocTemplate|None:
        template_types = DocTemplateFactory.__docTemplate_subclasses()
        
        find_type = [t() for t in template_types if type==t().get_id_signature()]
        return find_type[0] if find_type else None   



# Auteur : Esteban COLLARD, Nordine EL AMMARI
# Modifications : Reda ID TALEB & Manal LAGHMICH

from setuptools import find_packages, setup
import os.path


def recursive_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            if not filename.endswith(".pyc"):
                paths.append(os.path.join('..', path, filename))
    return paths

packages = find_packages()

def get_packages_data(packages: list[str]=packages):
    py_packs = dict([(p, ["*.py"]) for p in packages])
    unpackaged = {"": recursive_files("thonnycontrib/i18n/locale") + recursive_files("thonnycontrib/img")}
    return {**py_packs, **unpackaged} 

setupdir = os.path.dirname(__file__)


requirements = []
for line in open(os.path.join(setupdir, "requirements.txt"), encoding="ASCII"):
    if line.strip() and not line.startswith("#"):
        requirements.append(line)
        
setup(
    name="bimbam",
    version="6.998",
    author="idtaleb",
    description="A plug-in which adds a test framework",
    long_description="""A plug-in for Thonny which allows you to test your doc examples
 
More info: https://gitlab.univ-lille.fr/mirabelle.nebut/thonny-tests""",
    url="https://gitlab.univ-lille.fr/mirabelle.nebut/thonny-tests",
#    keywords="IDE education programming tests in documentation",
    classifiers=[
        "Topic :: Education :: Testing",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education"
        ],
    platforms=["Windows", "macOS", "Linux"],
    python_requires=">=3.9",
    package_data=get_packages_data(),
    install_requires=requirements,
    packages=packages,
)

